function hessian_mat = hessian(func,x,varargin)

% function hessian_mat = hessian(func,x,varargin)
% Computes second order partial derivatives
%
% INPUTS
%    func:           name of the function
%    x:              vector of variables around which the Hessian is calculated
%    varargin:       list of arguments following x
%
% OUTPUTS
%    hessian_matrix: Hessian matrix
%
% ALGORITHM
% Uses Abramowitz and Stegun (1965) formulas 25.3.24 and 25.3.27 p. 884
%
% SPECIAL REQUIREMENTS
%    none
%  

% Copyright (C) 2001-2007 Dynare Team
%
% This file is part of Dynare.
%
% Dynare is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% Dynare is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with Dynare.  If not, see <http://www.gnu.org/licenses/>.

global options_

% func = str2func(func);


n=size(x,1);
%h1=max(abs(x),options_.gstep*ones(n,1))*eps^(1/3);
h1=max(abs(x),sqrt(options_.gstep)*ones(n,1))*eps^(1/6);
h_1=h1;
xh1=x+h1;
h1=xh1-x;
xh1=x-h_1;
h_1=x-xh1;
f0=feval(func,x,varargin{:});
f1=zeros(n,1);
f_1=f1;


func = [func, '_parallel'];
global bayestopt_ estim_params_ trend_coeff_ M_ oo_ xparam1_test ...
			it_;
pbayestopt_ = bayestopt_;
pestim_params_ = estim_params_;
poptions_ = options_;
ptrend_coeff_ = trend_coeff_;
pM_ = M_;
poo_ = oo_;
pxparam1_test = xparam1_test;
pit_ = it_;

parfor i=1:n
    xh1 = x;
    xh1(i) = x(i) + h1(i);
    f1(i) = feval(func,xh1, ...
					pbayestopt_,  pestim_params_,  poptions_,  ptrend_coeff_,  pM_,  poo_,  pxparam1_test,  pit_, ...
					varargin{:});
    xh1(i) = x(i) - h_1(i);
    f_1(i) = feval(func,xh1, ...
					pbayestopt_,  pestim_params_,  poptions_,  ptrend_coeff_,  pM_,  poo_,  pxparam1_test,  pit_, ...
					varargin{:});
end


h_diag = (f1 + f_1 - 2*f0)./(h1.*h_1);
h_cols = zeros(n,n);
parfor i=1:n
    thiscol = zeros(n,1);
    this_h_1 = h_1;
    for j=i+1:n
        xh_1 = x;
        xh1 = x;
        temp = f1 + f_1 - f0*ones(n,1);
        xh1(i) = x(i) + h1(i);
        xh1(j) = x(j) + this_h_1(j);
        xh_1(i) = x(i) - h1(i);
        xh_1(j) = x(j) - this_h_1(j);
        thiscol(j) =-(-feval(func,xh1, ...
			pbayestopt_,  pestim_params_,  poptions_,  ptrend_coeff_,  pM_,  poo_,  pxparam1_test,  pit_, ...
						varargin{:})-feval(func,xh_1, ...
			pbayestopt_,  pestim_params_,  poptions_,  ptrend_coeff_,  pM_,  poo_,  pxparam1_test,  pit_, ...
						varargin{:})+temp(i)+temp(j))./(2*h1(i)*h_1(j));
    end
    h_cols(i,:) = thiscol;
end

hessian_mat = diag(h_diag) + h_cols + h_cols';


