function [xparam1, hh, gg, fval, igg] = newrat(func0, x, hh, gg, igg, ftol0, nit, flagg, varargin)
%
%  Copyright (C) 2004 Marco Ratto
%
%  [xparam1, hh, gg, fval, igg] = newrat(func0, x, hh, gg, igg, ftol0, nit, flagg, varargin)
%
%  Optimiser with outer product gradient and 'Gibbs type' steps
%  uses Chris Sims subroutine for line search
%
%  func0 = name of the function
%  there must be a version of the function called [func0,'_hh.m'], that also
%  gives as second OUTPUT the single contributions at times t=1,...,T
%    of the log-likelihood to compute outer product gradient
%
%  x = starting guess
%  hh = initial Hessian [OPTIONAL]
%  gg = initial gradient [OPTIONAL]
%  igg = initial inverse Hessian [OPTIONAL]
%  ftol0 = ending criterion for function change 
%  nit = maximum number of iterations
%
%  In each iteration, Hessian is computed with outer product gradient.
%  for final Hessian (to start Metropolis):
%  flagg = 0, final Hessian computed with outer product gradient
%  flagg = 1, final 'mixed' Hessian: diagonal elements computed with numerical second order derivatives
%             with correlation structure as from outer product gradient, 
%  flagg = 2, full numerical Hessian
%
%  varargin = list of parameters for func0 
%

global bayestopt_ estim_params_ options_ trend_coeff_ M_ oo_ xparam1_test ...
			it_;
  
icount=0;
nx=length(x);
xparam1=x;
%ftol0=1.e-6;
htol_base = max(1.e-5, ftol0);
flagit=0;  % mode of computation of hessian in each iteration
ftol=ftol0;
gtol=1.e-3;
htol=htol_base;
htol0=htol_base;
gibbstol=length(bayestopt_.pshape)/50; %25;

func_hh = [func0,'_hh'];
func = str2func(func0);
fval0=feval(func,x,varargin{:});
fval=fval0;
if isempty(hh)
    [dum, gg, htol0, igg, hhg]=mr_hessian(func_hh,x,flagit,htol,varargin{:});
    hh0 = reshape(dum,nx,nx);
    hh=hhg;
    if min(eig(hh0))<0,
        hh0=hhg; %generalized_cholesky(hh0);
    elseif flagit==2,
        hh=hh0;
        igg=inv(hh);
    end
    if htol0>htol,
        htol=htol0;
        %ftol=htol0;
    end
else
    hh0=hh;
    hhg=hh;
    igg=inv(hh);
end
disp(['Gradient norm ',num2str(norm(gg))])
ee=eig(hh);
disp(['Minimum Hessian eigenvalue ',num2str(min(ee))])
disp(['Maximum Hessian eigenvalue ',num2str(max(ee))])
g=gg;
check=0;
if max(eig(hh))<0, disp('Negative definite Hessian! Local maximum!'), pause, end,
save m1 x hh g hhg igg fval0

igrad= ones(4,1);
igibbs = ones(4,1);
inx=eye(nx);
jit=0;
nig=cell(4,1);
ig=ones(nx,4);


bigH = repmat(igg, [1 1 4]);


pfcn = [func0, '_parallel'];
pbayestopt_ = bayestopt_;
pestim_params_ = estim_params_;
poptions_ = options_;
ptrend_coeff_ = trend_coeff_;
pM_ = M_;
poo_ = oo_;
pxparam1_test = xparam1_test;
pit_ = it_;


while norm(gg)>gtol & check==0 & jit<nit,
	
    jit=jit+1;
	
    tic
	
    icount=icount+1;
    bayestopt_.penalty = fval0(icount);
    disp([' '])
    disp(['Iteration ',num2str(icount)])
	
	fval = zeros(4,1);
	x0 = cell(4,1);
	
	method_name = {'BFGS', 'Biggs', 'SR1', 'MR Hessian'};
	
	gstep_ = options_.gstep;
	h2=bayestopt_.ub-bayestopt_.lb;
	
 	parfor i = 1:4

		
		igg = bigH(:,:,i);
		thisig = ig(:,i);
% 		thisx0 = x0{i};
		
		[fval(i) thisx0 fc retcode] = csminit(pfcn,xparam1,fval0(icount),gg,0,igg, ...
				pbayestopt_,  pestim_params_,  poptions_,  ptrend_coeff_,  pM_,  poo_,  pxparam1_test,  pit_, ...
				varargin{:});

		if igrad(i),
			[fval1 x01 fc retcode1] = csminit(pfcn,thisx0,fval(i),gg,0,inx, ...
				pbayestopt_,  pestim_params_,  poptions_,  ptrend_coeff_,  pM_,  poo_,  pxparam1_test,  pit_, ...
				varargin{:});
			if (fval(i)-fval1)>1, %(fval0(icount)-fval),
				fprintf('%s: gradient step!\n', method_name{i});
			else
				igrad(i)=0;
			end
			fval(i) = fval1;
			thisx0 = x01;
			fprintf('%s: improvement of %f\n', method_name{i}, fval0(icount)-fval(i));
		end

		if (fval0(icount)-fval(i))<1.e-2*(gg'*(igg*gg))/2 & igibbs(i),
			if length(find(thisig))<nx,
				ggx=zeros(nx,1);
				ggx(find(thisig))=gg(find(thisig));
				hhx = reshape(dum,nx,nx);
				iggx=eye(length(gg));
				iggx(find(thisig),find(thisig)) = inv( hhx(find(thisig),find(thisig)) );
				[fvala thisx0 fc retcode] = csminit(pfcn,thisx0,fval(i),ggx,0,iggx, ...
				pbayestopt_,  pestim_params_,  poptions_,  ptrend_coeff_,  pM_,  poo_,  pxparam1_test,  pit_, ...
				varargin{:});
			end
			h1=max(abs(thisx0),sqrt(gstep_)*ones(length(thisx0),1))*eps^(1/4);

			[fvala, thisx0, thisig] = mr_gstep_parallel(pfcn,thisx0,htol,h1, h2, method_name{i},  ...
				pbayestopt_,  pestim_params_,  poptions_,  ptrend_coeff_,  pM_,  poo_,  pxparam1_test,  pit_, ...
				varargin{:});
			nig{i}=[nig{i} thisig];
			 if (fval(i)-fvala)<gibbstol*(fval0(icount)-fval(i)),
				 igibbs(i)=0;
				 fprintf('%s: no more Gibbs steps.\n', method_name{i});
			 else
				fprintf('%s: Hooray for the Gibbs step!\n', method_name{i});
			end
			fval(i)=fvala;
			fprintf('%s: improvement of %f\n', method_name{i}, fval0(icount)-fval(i));
		end
		if (fval0(icount)-fval(i))<ftol & flagit==0,
			fprintf('%s: diagonal Hessian\n', method_name{i});
			ihh=diag(1./( abs(diag(igg)) ));        
			[fval2 thisx0 fc retcode2] = csminit(pfcn,thisx0,fval(i),gg,0,ihh, ...
				pbayestopt_,  pestim_params_,  poptions_,  ptrend_coeff_,  pM_,  poo_,  pxparam1_test,  pit_, ...
					varargin{:});
% 				if (fval(i)-fval2)>=ftol ,
% 					%hh=diag(diag(hh));
% 					disp('Diagonal Hessian successful')            
% 				end
			fval(i)=fval2;
			fprintf('%s: improvement of %f\n', method_name{i}, fval0(icount)-fval(i));
		end        
		if (fval0(icount)-fval(i))<ftol & flagit==0,
			fprintf('%s: gradient direction\n', method_name{i});
			ihh0=inx.*1.e-4;        
			[fval3 thisx0 fc retcode3] = csminit(pfcn,thisx0,fval(i),gg,0,ihh0,  ...
				pbayestopt_,  pestim_params_,  poptions_,  ptrend_coeff_,  pM_,  poo_,  pxparam1_test,  pit_, ...
					varargin{:});
				if (fval(i)-fval3)>=ftol ,
					%hh=hh0;
					%ihh=ihh0;
					disp('Gradient direction successful')            
				end
				fval(i)=fval3;
			fprintf('%s: improvement of %f\n', method_name{i}, fval0(icount)-fval(i));
		end
		
		ig(:,i) = thisig;
		x0{i} = thisx0;
	end
	
	igg = bigH(:,:,1);		% Use the BFGS hessian as the default
	
	[bestf ix] = min(fval);
	pickstr = {'','','',''};
	pickstr{ix} = '*';
	
	xold = xparam1;
    xparam1 = x0{ix};
	fprintf('Results from 4-way search:\n');
	fprintf('BFGS%s: %f, Biggs%s: %f, SR1%s: %f, MR Hessian%s: %f\n\n', ...
			pickstr{1}, fval(1), pickstr{2}, fval(2), pickstr{3}, fval(3), pickstr{4}, fval(4));
	
	fval = fval(ix);
	
    x(:,icount+1)=xparam1;
    fval0(icount+1)=fval;
    %if (fval0(icount)-fval)<ftol*ftol & flagg==1;,
    if (fval0(icount)-fval)<ftol,
        disp('No further improvement is possible!')
        check=1;
        if flagit==2,
            hh=hh0;
        elseif flagg>0,
            [dum, gg, htol0, igg, hhg]=mr_hessian(func_hh,xparam1,flagg,ftol0,varargin{:});   
            if flagg==2,
                hh = reshape(dum,nx,nx);
                ee=eig(hh);
                if min(ee)<0
                    hh=hhg;
                end
            else
                hh=hhg;
            end
        end
        disp(['Actual dxnorm ',num2str(norm(x(:,end)-x(:,end-1)))])
        disp(['FVAL          ',num2str(fval)])
        disp(['Improvement   ',num2str(fval0(icount)-fval)])
        disp(['Ftol          ',num2str(ftol)])
        disp(['Htol          ',num2str(htol0)])
        disp(['Gradient norm  ',num2str(norm(gg))])
        ee=eig(hh);
        disp(['Minimum Hessian eigenvalue ',num2str(min(ee))])
        disp(['Maximum Hessian eigenvalue ',num2str(max(ee))])
         g(:,icount+1)=gg;
    else
        
        df = fval0(icount)-fval;
        disp(['Actual dxnorm ',num2str(norm(x(:,end)-x(:,end-1)))])
        disp(['FVAL          ',num2str(fval)])
        disp(['Improvement   ',num2str(df)])
        disp(['Ftol          ',num2str(ftol)])
        disp(['Htol          ',num2str(htol0)])

        if df<htol0,
            htol=max(htol_base,df/10);
        end
        
        if norm(x(:,icount)-xparam1)>1.e-12,
            save m1 x fval0 nig -append


			oldgg = gg;
			gg = numgrad(func0, xparam1, varargin{:});
			delta = xparam1 - xold;
			y = gg - oldgg;
			if size(y,2)>1
				y=y';
			end
			if size(delta,2)>1
				delta = delta';
			end
			
			
			% BFGS update + Biggs update
			if (abs(delta'*y) >1e-12)
% 				bigH(:,:,1) = igg + ( (1+ y'*igg*y/(delta'*y))*delta*delta' - delta*y'*igg - igg*y*delta'  )/(delta'*y);
				bigH(:,:,1) = bigH(:,:,1) + ( (1+ y'*igg*y/(delta'*y))*delta*delta' - delta*y'*igg - igg*y*delta'  )/(delta'*y);
				
				biggsT = (6/(delta'*y))*( df + delta'*gg ) - 2;
				biggsLambda = 1/biggsT;
% 				bigH(:,:,2) = igg + ( (biggsLambda + y'*igg*y/(delta'*y))*delta*delta' - delta*y'*igg - igg*y*delta' )/(delta'*y);
				bigH(:,:,2) = bigH(:,:,2) + ( (biggsLambda + y'*igg*y/(delta'*y))*delta*delta' - delta*y'*igg - igg*y*delta' )/(delta'*y);
			else
			   disp('bfgs update failed.')
			   disp(['|dg| = ' num2str(sqrt(dg'*dg)) '|dx| = ' num2str(sqrt(dx'*dx))]);
			   disp(['dg''*dx = ' num2str(dgdx)])
			   disp(['|H*dg| = ' num2str(Hdg'*Hdg)])
			end

			% Symmetric Rank One update
			denom = (delta - igg*y)'*y;
			if(abs(denom) > 1e-12)
% 				bigH(:,:,3) = igg + (delta - igg*y)*(delta - igg*y)'/denom;
				bigH(:,:,3) = bigH(:,:,3) + (delta - igg*y)*(delta - igg*y)'/denom;
			else
				disp('SR1 update failed.');
			end
			
			
			[dum, gg2, htol0, igg, hhg]=mr_hessian(func_hh,xparam1,flagit,htol,varargin{:});
			if htol0>htol, %ftol,
				%ftol=htol0;
				htol=htol0;
				disp(' ')
				disp('Numerical noise in the likelihood')
				disp('Tolerance has to be relaxed')
				disp(' ')
%             elseif htol0<ftol,
%                 ftol=max(htol0, ftol0);
			end
			hh0 = reshape(dum,nx,nx);
			hh=hhg;
			if flagit==2,
				if min(eig(hh0))<=0,
					hh0=hhg; %generalized_cholesky(hh0);
				else 
					hh=hh0;
					igg=inv(hh);
				end
			end
			bigH(:,:,4) = igg;
			
			
			if( (norm(-bigH(:,:,1)*gg) > 1e12) || (norm(-bigH(:,:,2)*gg) > 1e12) || (norm(-bigH(:,:,3)*gg) > 1e12) )
				disp('Near-singular H problem for at least one quasi-Newton method;');
				disp('I will use MR Hessian for the next q-N iteration.')
				
				bigH(:,:,1) = igg;
				bigH(:,:,2) = igg;
				bigH(:,:,3) = igg;
			end
			
        end
        disp(['Gradient norm  ',num2str(norm(gg))])
		
		
%         ee=eig(hh);
		ee(:,1) = eig(bigH(:,:,1));
		ee(:,2) = eig(bigH(:,:,2));
		ee(:,3) = eig(bigH(:,:,3));
		
        disp(['Minimum Hessian eigenvalue ',num2str(min(min(ee)))])
        disp(['Maximum Hessian eigenvalue ',num2str(max(max(ee)))])
        if max(eig(hh))<0, disp('Negative definite Hessian! Local maximum!'), pause, end,
        t=toc;
        disp(['Elapsed time for iteration ',num2str(t),' s.'])
        
         g(:,icount+1)=gg;
        save m1 x hh g hhg igg fval0 nig
    end
end


save m1 x hh g hhg igg fval0 nig
if ftol>ftol0,
    disp(' ')
    disp('Numerical noise in the likelihood')
    disp('Tolerance had to be relaxed')
    disp(' ')
end

if jit==nit,
    disp(' ')
    disp('Maximum number of iterations reached')
    disp(' ')
end

if norm(gg)<=gtol,
    disp(['Estimation ended:'])
    disp(['Gradient norm < ', num2str(gtol)])
end
if check==1,
    disp(['Estimation successful.'])
end

return

%  
function f00 = lsearch(lam,func,x,dx,varargin)


x0=x-dx*lam;
f00=feval(func,x0,varargin{:});






